import openapi_spec_validator as osv
import pytest
from apispec import BasePlugin
from flask import Blueprint
from flask.views import MethodView

from .schemas import Bar
from .schemas import Foo
from .schemas import Pagination
from apiflask import APIBlueprint
from apiflask import APIFlask
from apiflask import Schema
from apiflask.exceptions import abort
from apiflask.fields import Integer
from apiflask.fields import String


def test_app_init(app):
    assert app
    assert hasattr(app, 'import_name')
    assert hasattr(app, 'title')
    assert 'TAGS' in app.config
    assert 'openapi' in app.blueprints


def test_json_errors(app, client):
    assert app.json_errors is True

    rv = client.get('/not-exist')
    assert rv.status_code == 404
    assert rv.headers['Content-Type'] == 'application/json'
    assert 'message' in rv.json
    assert 'detail' in rv.json
    assert rv.json['message'] == 'Not Found'
    assert rv.json['detail'] == {}

    app = APIFlask(__name__, json_errors=False)
    assert app.json_errors is False
    rv = app.test_client().get('/not-exist')
    assert rv.status_code == 404
    assert rv.headers['Content-Type'] == 'text/html; charset=utf-8'
    assert b'!doctype' in rv.data.lower()


def test_json_errors_reuse_werkzeug_headers(app, client):
    # test automatic 405
    @app.get('/foo')
    def foo():
        pass

    # test manually raise 405
    @app.get('/bar')
    def bar():
        from werkzeug.exceptions import MethodNotAllowed

        raise MethodNotAllowed(valid_methods=['GET'])

    rv = client.post('/foo')
    assert rv.status_code == 405
    assert 'Allow' in rv.headers

    rv = client.get('/bar')
    assert rv.status_code == 405
    assert rv.headers['Content-Type'] == 'application/json'
    assert rv.headers['Allow'] == 'GET'


def test_view_function_arguments_order(app, client):
    class Query(Schema):
        foo = String(required=True)
        bar = String(required=True)

    class Pet(Schema):
        name = String(required=True)
        age = Integer(dump_default=123)

    @app.post('/pets/<int:pet_id>/toys/<int:toy_id>')
    @app.input(Query, location='query', arg_name='query')
    @app.input(Pagination, location='query', arg_name='pagination')
    @app.input(Pet, location='json')
    def pets(pet_id, toy_id, query, pagination, json_data):
        return {
            'pet_id': pet_id,
            'toy_id': toy_id,
            'foo': query['foo'],
            'bar': query['bar'],
            'pagination': pagination,
            **json_data,
        }

    @app.route('/animals/<int:pet_id>/toys/<int:toy_id>')
    class Animals(MethodView):
        @app.input(Query, location='query', arg_name='query')
        @app.input(Pagination, location='query', arg_name='pagination')
        @app.input(Pet, location='json')
        def post(self, pet_id, toy_id, query, pagination, json_data):
            return {
                'pet_id': pet_id,
                'toy_id': toy_id,
                'foo': query['foo'],
                'bar': query['bar'],
                'pagination': pagination,
                **json_data,
            }

    rv = client.post('/pets/1/toys/3?foo=yes&bar=no', json={'name': 'dodge', 'age': 5})
    assert rv.status_code == 200
    assert rv.json['pet_id'] == 1
    assert rv.json['toy_id'] == 3
    assert rv.json['foo'] == 'yes'
    assert rv.json['bar'] == 'no'
    assert rv.json['name'] == 'dodge'
    assert rv.json['age'] == 5
    assert rv.json['pagination']['page'] == 1
    assert rv.json['pagination']['per_page'] == 10

    rv = client.post('/animals/1/toys/3?foo=yes&bar=no', json={'name': 'dodge', 'age': 5})
    assert rv.status_code == 200
    assert rv.json['pet_id'] == 1
    assert rv.json['toy_id'] == 3
    assert rv.json['foo'] == 'yes'
    assert rv.json['bar'] == 'no'
    assert rv.json['name'] == 'dodge'
    assert rv.json['age'] == 5
    assert rv.json['pagination']['page'] == 1
    assert rv.json['pagination']['per_page'] == 10


def test_error_callback(app, client):
    @app.error_processor
    def custom_error_handler(e):
        assert e.status_code == 500
        assert e.detail == {}
        assert e.headers == {}
        assert e.message == 'Internal Server Error'
        return {'message': 'something was wrong'}, 200

    @app.get('/')
    def error():
        print(a)  # noqa: F821
        return ''

    rv = client.get('/not-exist')
    assert rv.status_code == 200
    assert 'message' in rv.json
    assert rv.json['message'] == 'something was wrong'

    rv = client.get('/')
    assert rv.status_code == 200
    assert 'message' in rv.json
    assert rv.json['message'] == 'something was wrong'


def test_apply_custom_error_callback_for_generic_errors():
    app = APIFlask(__name__, json_errors=False)
    client = app.test_client()

    @app.error_processor
    def custom_error_handler(e):
        return {'message': 'custom handler'}, e.status_code

    rv = client.get('/not-exist')
    assert rv.status_code == 404
    assert 'message' in rv.json
    assert rv.json['message'] == 'custom handler'


def test_skip_raw_blueprint(app, client):
    raw_bp = Blueprint('raw', __name__)
    api_bp = APIBlueprint('api', __name__, tag='test')

    @raw_bp.route('/foo')
    def foo():
        pass

    @raw_bp.route('/bar')
    class Bar(MethodView):
        def get(self):
            pass

    @api_bp.get('/baz')
    def baz():
        pass

    @api_bp.route('/spam')
    class Spam(MethodView):
        def get(self):
            pass

    app.register_blueprint(raw_bp)
    app.register_blueprint(api_bp)

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['tags'] == [{'name': 'test'}]
    assert '/foo' not in rv.json['paths']
    assert '/bar' not in rv.json['paths']
    assert '/baz' in rv.json['paths']
    assert '/spam' in rv.json['paths']


def test_dispatch_static_request(app, client):
    # positional arguments
    @app.get('/mystatic/<int:pet_id>')
    @app.input(Foo)
    def mystatic(pet_id, json_data):  # endpoint: mystatic
        return {'pet_id': pet_id, 'foo': json_data}

    # positional arguments
    # blueprint static route accepts both keyword/positional arguments
    bp = APIBlueprint('foo', __name__, static_folder='static')
    app.register_blueprint(bp, url_prefix='/foo')

    rv = client.get('/mystatic/2', json={'id': 1, 'name': 'foo'})
    assert rv.status_code == 200
    assert rv.json['pet_id'] == 2
    assert rv.json['foo'] == {'id': 1, 'name': 'foo'}

    rv = client.get('/foo/static/hello.css')  # endpoint: foo.static
    assert rv.status_code == 404

    # keyword arguments
    rv = client.get('/static/hello.css')  # endpoint: static
    assert rv.status_code == 404


def schema_name_resolver1(schema):
    name = schema.__class__.__name__
    if schema.partial:
        name += '_'
    return name


def schema_name_resolver2(schema):
    name = schema.__class__.__name__
    if name.endswith('Schema'):
        name = name[:-6] or name
    if schema.partial:
        name += 'Partial'
    return name


@pytest.mark.parametrize(
    'resolver', [APIFlask._schema_name_resolver, schema_name_resolver1, schema_name_resolver2]
)
def test_schema_name_resolver(app, client, resolver):
    app.schema_name_resolver = resolver

    @app.route('/foo')
    @app.output(Foo)
    def foo():
        pass

    @app.route('/bar')
    @app.output(Bar(partial=True))
    def bar():
        pass

    spec = app.spec
    if resolver == schema_name_resolver1:
        assert 'Foo' in spec['components']['schemas']
        assert 'Bar_' in spec['components']['schemas']
    elif resolver == schema_name_resolver2:
        assert 'Foo' in spec['components']['schemas']
        assert 'BarPartial' in spec['components']['schemas']
    else:
        assert 'Foo' in spec['components']['schemas']
        assert 'BarUpdate' in spec['components']['schemas']


@pytest.mark.parametrize('ui_name', ['swagger-ui', 'redoc', 'elements', 'rapidoc', 'rapipdf'])
def test_docs_ui(ui_name):
    app = APIFlask(__name__, docs_ui=ui_name)
    client = app.test_client()

    rv = client.get('/docs')
    assert rv.status_code == 200


def test_bad_docs_ui():
    with pytest.raises(ValueError):
        APIFlask(__name__, docs_ui='bad')


def test_return_list_as_json(app, client):
    test_list = ['foo', 'bar', 'baz']

    @app.get('/single')
    def single_value():
        return test_list

    @app.get('/multi')
    def multi_values():
        return test_list, 201, {'X-Foo': 'bar'}

    rv = client.get('/single')
    assert rv.status_code == 200
    assert rv.headers['Content-Type'] == 'application/json'
    assert rv.json == test_list

    rv = client.get('/multi')
    assert rv.status_code == 201
    assert rv.headers['Content-Type'] == 'application/json'
    assert rv.json == test_list


def test_apispec_plugins(app):
    class TestPlugin(BasePlugin):
        def operation_helper(self, path=None, operations=None, **kwargs) -> None:
            operations.update({'post': 'some_injected_test_data'})

    app.spec_plugins = [TestPlugin()]

    @app.get('/plugin_test')
    def single_value():
        return 'plugin_test'

    spec = app._get_spec('json')

    assert spec['paths']['/plugin_test'].get('post') == 'some_injected_test_data'


def test_spec_decorators(app, client):
    def auth_decorator(f):
        def wrapper(*args, **kwargs):
            abort(401)
            return f(*args, **kwargs)

        return wrapper

    app.config['SPEC_DECORATORS'] = [auth_decorator]

    rv = client.get('/openapi.json')
    assert rv.status_code == 401

    rv = client.get('/docs')
    assert rv.status_code == 200


def test_docs_decorators(app, client):
    def auth_decorator(f):
        def wrapper(*args, **kwargs):
            abort(401)
            return f(*args, **kwargs)

        return wrapper

    app.config['DOCS_DECORATORS'] = [auth_decorator]

    rv = client.get('/openapi.json')
    assert rv.status_code == 200

    rv = client.get('/docs')
    assert rv.status_code == 401


def test_swagger_ui_oauth_redirect_decorators(app, client):
    def auth_decorator(f):
        def wrapper(*args, **kwargs):
            abort(401)
            return f(*args, **kwargs)

        return wrapper

    app.config['SWAGGER_UI_OAUTH_REDIRECT_DECORATORS'] = [auth_decorator]

    rv = client.get('/openapi.json')
    assert rv.status_code == 200

    rv = client.get('/docs')
    assert rv.status_code == 200

    rv = client.get('/docs/oauth2-redirect')
    assert rv.status_code == 401



import openapi_spec_validator as osv

from .schemas import DelimitedListQuery


def test_query_delimited_list_support(app, client):
    @app.route('/users')
    @app.input(DelimitedListQuery, location='query')
    def get_users(query_data):
        """Get Users"""
        return query_data['tags']

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/users']['get']['parameters'][0]['name'] == 'tags'
    assert rv.json['paths']['/users']['get']['parameters'][0]['style'] == 'form'
    assert rv.json['paths']['/users']['get']['parameters'][0]['explode'] is False

    rv = client.get('/users?tags=admin,editor,user')
    assert rv.status_code == 200
    assert rv.json == ['admin', 'editor', 'user']



import openapi_spec_validator as osv
import pytest
from flask.views import MethodView
from flask.views import View

from .schemas import Foo
from apiflask import APIBlueprint
from apiflask import HTTPTokenAuth


@pytest.mark.parametrize('method', ['get', 'post', 'put', 'patch', 'delete'])
def test_route_shortcuts(app, client, method):
    route_method = getattr(app, method)
    client_method = getattr(client, method)

    @route_method('/pet')
    @app.output(Foo)
    def test_shortcuts():
        return {'name': method}

    assert client_method('/pet').json['name'] == method

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/pet'][method]


def test_route_on_method_view(app, client):
    @app.route('/')
    class Foo(MethodView):
        def get(self):
            return 'get'

        def post(self):
            return 'post'

        def delete(self):
            return 'delete'

        def put(self):
            return 'put'

        def patch(self):
            return 'patch'

    rv = client.get('/')
    assert rv.data == b'get'
    rv = client.post('/')
    assert rv.data == b'post'
    rv = client.delete('/')
    assert rv.data == b'delete'
    rv = client.put('/')
    assert rv.data == b'put'
    rv = client.patch('/')
    assert rv.data == b'patch'


def test_blueprint_route_on_method_view(app, client):
    bp = APIBlueprint('test', __name__, tag='foo')

    @bp.route('/')
    class Foo(MethodView):
        def get(self):
            return 'get'

        def post(self):
            return 'post'

        def delete(self):
            return 'delete'

        def put(self):
            return 'put'

        def patch(self):
            return 'patch'

    app.register_blueprint(bp)

    rv = client.get('/')
    assert rv.data == b'get'
    rv = client.post('/')
    assert rv.data == b'post'
    rv = client.delete('/')
    assert rv.data == b'delete'
    rv = client.put('/')
    assert rv.data == b'put'
    rv = client.patch('/')
    assert rv.data == b'patch'


def test_bad_route_decorator_usages(app):
    with pytest.raises(RuntimeError):

        @app.get('/foo', methods=['GET'])
        def foo(self):
            pass

    with pytest.raises(RuntimeError):

        @app.get('/baz')
        class Baz(MethodView):
            def get(self):
                pass


def test_class_attribute_decorators(app, client):
    auth = HTTPTokenAuth()

    @app.route('/')
    class Foo(MethodView):
        decorators = [app.auth_required(auth), app.doc(responses=[404])]

        def get(self):
            pass

        def post(self):
            pass

    rv = client.get('/')
    assert rv.status_code == 401
    rv = client.post('/')
    assert rv.status_code == 401

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert '404' in rv.json['paths']['/']['get']['responses']
    assert '404' in rv.json['paths']['/']['post']['responses']
    assert 'BearerAuth' in rv.json['paths']['/']['get']['security'][0]
    assert 'BearerAuth' in rv.json['paths']['/']['post']['security'][0]


def test_overwrite_class_attribute_decorators(app, client):
    @app.route('/')
    class Foo(MethodView):
        decorators = [app.doc(deprecated=True, tags=['foo'])]

        def get(self):
            pass

        @app.doc(deprecated=False)
        def post(self):
            pass

        @app.doc(tags=['bar'])
        def delete(self):
            pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/']['get']['deprecated']
    assert rv.json['paths']['/']['get']['tags'] == ['foo']
    assert rv.json['paths']['/']['post']['tags'] == ['foo']
    assert rv.json['paths']['/']['delete']['tags'] == ['bar']
    assert 'deprecated' not in rv.json['paths']['/']['post']


def test_add_url_rule_with_method_view(app, client):
    class Foo(MethodView):
        def get(self):
            return 'get'

        def post(self):
            """Create foo"""
            return 'post'

    app.add_url_rule('/', view_func=Foo.as_view('foo'), methods=['GET', 'POST'])

    rv = client.get('/')
    assert rv.data == b'get'
    rv = client.post('/')
    assert rv.data == b'post'

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/']['get']['summary'] == 'Get Foo'
    assert rv.json['paths']['/']['post']['summary'] == 'Create foo'


def test_add_url_rule_with_method_view_as_view(app, client):
    class Foo(MethodView):
        def get(self):
            return 'get'

        def post(self):
            """Create foo"""
            return 'post'

    foo_view = Foo.as_view('foo')
    app.add_url_rule('/foo/get', view_func=foo_view, methods=['GET'])
    app.add_url_rule('/foo/post', view_func=foo_view, methods=['POST'])

    rv = client.get('/foo/get')
    assert rv.data == b'get'
    rv = client.post('/foo/post')
    assert rv.data == b'post'

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/foo/get']['get']['summary'] == 'Get Foo'
    assert rv.json['paths']['/foo/post']['post']['summary'] == 'Create foo'


def test_view_endpoint_contains_dot(app, client):
    @app.route('/', endpoint='hello.world')
    def foo():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/']['get']


@pytest.mark.parametrize('view_class', [View])
def test_add_url_rule_skip_collecting_spec_from_view_class(app, client, view_class):
    class Foo(view_class):
        def dispatch_request(self):
            return 'Hello'

    app.add_url_rule('/foo', view_func=Foo.as_view('foo'))

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    assert '/foo' not in rv.json['paths']

    rv = client.get('/foo')
    assert rv.status_code == 200


def test_empty_methodview_class(app, client):
    class Foo(MethodView):
        pass

    app.add_url_rule('/foo', view_func=Foo.as_view('foo'))
    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    assert '/foo' not in rv.json['paths']


def test_methodview_class_with_subclass(app, client):
    class Foo(MethodView):
        text = 'You are in Foo'

        def get(self):
            return self.text

    class FooSubClass(Foo):
        text = 'You are in FooSubClass'

    app.add_url_rule('/foo', view_func=FooSubClass.as_view('foo'))

    rv = client.get('/foo')
    assert rv.status_code == 200
    assert rv.data == b'You are in FooSubClass'

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/foo']['get']['summary'] == 'Get FooSubClass'



from marshmallow import EXCLUDE
from marshmallow.validate import Range

from apiflask import Schema
from apiflask.fields import Decimal
from apiflask.fields import DelimitedList
from apiflask.fields import File
from apiflask.fields import Integer
from apiflask.fields import List
from apiflask.fields import String
from apiflask.validators import OneOf


class Foo(Schema):
    class Meta:
        unknown = EXCLUDE

    id = Integer(dump_default=123)
    name = String(required=True)


class Bar(Schema):
    class Meta:
        unknown = EXCLUDE

    id2 = Integer(dump_default=123)
    name2 = String(required=True)


class Baz(Schema):
    id = Integer(dump_default=123)
    name = String()


class Qux(Schema):
    number = Decimal(validate=Range(min=0))


class Query(Schema):
    class Meta:
        unknown = EXCLUDE

    id = Integer(load_default=1)


class Form(Schema):
    name = String()


class Files(Schema):
    image = File()


class FilesList(Schema):
    images = List(File())


class FormAndFiles(Schema):
    name = String()
    image = File()


class EnumPathParameter(Schema):
    image_type = String(validate=OneOf(['jpg', 'png', 'tiff', 'webp']))


class Pagination(Schema):
    class Meta:
        unknown = EXCLUDE

    page = Integer(load_default=1)
    per_page = Integer(load_default=10)


class Header(Schema):
    class Meta:
        unknown = EXCLUDE

    foo = String(load_default='bar')


class ValidationError(Schema):
    status_code = String(required=True)
    message = String(required=True)


class HTTPError(Schema):
    status_code = String(required=True)
    message = String(required=True)


class CustomHTTPError(Schema):
    status_code = String(required=True)
    message = String(required=True)
    custom = String(required=True)


class ResponseHeader(Schema):
    x_token = String(
        data_key='X-Token', required=True, metadata={'description': 'A custom token header'}
    )


class DelimitedListQuery(Schema):
    tags = DelimitedList(String)



import openapi_spec_validator as osv


def test_specification_extensions(app, client):
    @app.get('/')
    @app.doc(extensions={'x-foo': {'foo': 'bar'}})
    def foo():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/']['get']['x-foo'] == {'foo': 'bar'}



from apiflask import APIFlask, Schema, abort
from apiflask.fields import Integer, String, Field
from apiflask.validators import Length, OneOf

app = APIFlask(__name__)


class BaseResponse(Schema):
    data = Field()  # the data key
    message = String()
    code = Integer()


app.config['BASE_RESPONSE_SCHEMA'] = BaseResponse
# the data key should match the data field name in the base response schema
# defaults to "data"
app.config['BASE_RESPONSE_DATA_KEY'] = 'data'

pets = [
    {'id': 0, 'name': 'Kitty', 'category': 'cat'},
    {'id': 1, 'name': 'Coco', 'category': 'dog'},
    {'id': 2, 'name': 'Flash', 'category': 'cat'},
]


class PetIn(Schema):
    name = String(required=True, validate=Length(0, 10))
    category = String(required=True, validate=OneOf(['dog', 'cat']))


class PetOut(Schema):
    id = Integer()
    name = String()
    category = String()


@app.get('/')
def say_hello():
    data = {'message': 'Hello!'}
    return {'data': data, 'message': 'Success!', 'code': 200}


@app.get('/pets/<int:pet_id>')
@app.output(PetOut)
def get_pet(pet_id):
    if pet_id > len(pets) - 1 or pets[pet_id].get('deleted'):
        abort(404)
    return {
        'data': pets[pet_id],
        'message': 'Success!',
        'code': 200,
    }


@app.get('/pets')
@app.output(PetOut(many=True))
def get_pets():
    return {
        'data': pets,
        'message': 'Success!',
        'code': 200,
    }


@app.post('/pets')
@app.input(PetIn, location='json')
@app.output(PetOut, status_code=201)
def create_pet(json_data):
    pet_id = len(pets)
    json_data['id'] = pet_id
    pets.append(json_data)
    return {'data': pets[pet_id], 'message': 'Pet created.', 'code': 201}


@app.patch('/pets/<int:pet_id>')
@app.input(PetIn(partial=True), location='json')
@app.output(PetOut)
def update_pet(pet_id, json_data):
    if pet_id > len(pets) - 1:
        abort(404)
    for attr, value in json_data.items():
        pets[pet_id][attr] = value
    return {'data': pets[pet_id], 'message': 'Pet updated.', 'code': 200}


@app.delete('/pets/<int:pet_id>')
@app.output({}, status_code=204)
def delete_pet(pet_id):
    if pet_id > len(pets) - 1:
        abort(404)
    pets[pet_id]['deleted'] = True
    pets[pet_id]['name'] = 'Ghost'
    return ''



import openapi_spec_validator as osv
import pytest

from .schemas import Foo
from .schemas import ResponseHeader
from apiflask.fields import Boolean
from apiflask.fields import Integer
from apiflask.fields import List
from apiflask.fields import Number
from apiflask.fields import String


def test_spec_with_dict_headers(app, client):
    @app.route('/foo')
    @app.output(
        Foo,
        headers={
            'X-boolean': Boolean(metadata={'description': 'A boolean header'}),
            'X-integer': Integer(metadata={'description': 'An integer header'}),
            'X-number': Number(metadata={'description': 'A number header'}),
            'X-string': String(metadata={'description': 'A string header'}),
            'X-array': List(String(), metadata={'description': 'An array header'}),
        },
    )
    def foo():
        pass

    rv = client.get('/openapi.json')
    assert rv.json['paths']['/foo']['get']['responses']['200']['headers'] == {
        'X-boolean': {
            'description': 'A boolean header',
            'required': False,
            'schema': {'type': 'boolean'},
        },
        'X-integer': {
            'description': 'An integer header',
            'required': False,
            'schema': {'type': 'integer'},
        },
        'X-number': {
            'description': 'A number header',
            'required': False,
            'schema': {'type': 'number'},
        },
        'X-string': {
            'description': 'A string header',
            'required': False,
            'schema': {'type': 'string'},
        },
        'X-array': {
            'description': 'An array header',
            'required': False,
            'schema': {'items': {'type': 'string'}, 'type': 'array'},
            'style': 'form',
            'explode': True,
        },
    }


def test_spec_with_empty_headers(app, client):
    @app.route('/foo')
    @app.output(Foo, headers={})
    def foo():
        pass

    rv = client.get('/openapi.json')
    assert rv.json['paths']['/foo']['get']['responses']['200']['headers'] == {}


def test_spec_with_schema_headers(app, client):
    @app.route('/foo')
    @app.output(Foo, headers=ResponseHeader)
    def foo():
        pass

    rv = client.get('/openapi.json')
    assert rv.json['paths']['/foo']['get']['responses']['200']['headers'] == {
        'X-Token': {
            'description': 'A custom token header',
            'required': True,
            'schema': {'type': 'string'},
        },
    }


@pytest.mark.parametrize('openapi_version', ['3.0.0', '3.1.0'])
def test_spec_validity_with_headers(app, client, openapi_version):
    app.config['OPENAPI_VERSION'] = openapi_version

    @app.route('/foo')
    @app.output(Foo, headers=ResponseHeader)
    def foo():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)



import openapi_spec_validator as osv

from .schemas import Foo
from apiflask import HTTPTokenAuth


def test_async_view(app, client):
    @app.get('/')
    async def index():
        return {'message': 'hello'}

    rv = client.get('/')
    assert rv.status_code == 200
    assert rv.json['message'] == 'hello'


def test_async_error_processor(app, client):
    @app.error_processor
    async def custom_error_processor(e):
        return {'foo': 'test'}, e.status_code, e.headers

    rv = client.get('/foo')
    assert rv.status_code == 404
    assert rv.json['foo'] == 'test'


def test_async_spec_processor(app, client):
    @app.spec_processor
    async def update_spec(spec):
        spec['info']['title'] = 'Updated Title'
        return spec

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['info']['title'] == 'Updated Title'


def test_auth_required_on_async_view(app, client):
    auth = HTTPTokenAuth()

    @app.get('/')
    @app.auth_required(auth)
    async def index():
        pass

    rv = client.get('/')
    assert rv.status_code == 401


def test_doc_on_async_view(app, client):
    @app.get('/')
    @app.doc(summary='Test Root Endpoint')
    async def index():
        return {'echo': 1234}

    rv = client.get('/')
    assert rv.status_code == 200
    assert rv.json['echo'] == 1234


def test_input_on_async_view(app, client):
    @app.post('/')
    @app.input(Foo)
    async def index(json_data):
        return json_data

    data = {'id': 1, 'name': 'foo'}
    rv = client.post('/', json=data)
    assert rv.status_code == 200
    assert rv.json == data


def test_output_on_async_view(app, client):
    @app.get('/foo')
    @app.output(Foo)
    async def foo():
        return {'id': 1, 'name': 'foo'}

    rv = client.get('/foo')
    assert rv.status_code == 200
    assert rv.json['id'] == 1
    assert rv.json['name'] == 'foo'

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/foo']['get']['responses']['200']
    assert (
        rv.json['paths']['/foo']['get']['responses']['200']['content']['application/json'][
            'schema'
        ]['$ref']
        == '#/components/schemas/Foo'
    )


def test_async_doc_input_and_output_decorator(app, client):
    @app.post('/')
    @app.doc(summary='Test Root Endpoint')
    @app.input(Foo)
    @app.output(Foo)
    async def index(json_data):
        return json_data

    payload = {'id': 1, 'name': 'foo'}
    rv = client.post('/', json=payload)
    assert rv.status_code == 200
    assert rv.json == payload



from __future__ import annotations

import typing as t

from .base import SchemaAdapter

if t.TYPE_CHECKING:
    pass

try:
    from pydantic import BaseModel  # type: ignore[import-not-found]

    HAS_PYDANTIC = True
except ImportError:
    BaseModel = None  # type: ignore
    HAS_PYDANTIC = False

try:
    from marshmallow import Schema

    HAS_MARSHMALLOW = True
except ImportError:
    Schema = None  # type: ignore
    HAS_MARSHMALLOW = False


class SchemaRegistry:
    """Registry for managing schema adapters and automatic type detection.

    *Version added: 3.0.0*
    """

    def __init__(self) -> None:
        self._adapters: dict[str, type[SchemaAdapter]] = {}
        self._register_default_adapters()

    def _register_default_adapters(self) -> None:
        """Register the default schema adapters."""
        if HAS_MARSHMALLOW:
            from .marshmallow import MarshmallowAdapter

            self.register('marshmallow', MarshmallowAdapter)

        if HAS_PYDANTIC:
            from .pydantic import PydanticAdapter

            self.register('pydantic', PydanticAdapter)

    def register(self, name: str, adapter_class: type[SchemaAdapter]) -> None:
        """Register a schema adapter.

        Arguments:
            name: Name of the adapter
            adapter_class: Schema adapter class
        """
        self._adapters[name] = adapter_class

    def detect_schema_type(self, schema: t.Any) -> str:
        """Detect the type of schema.

        Arguments:
            schema: Schema object to detect type for

        Returns:
            Schema type name

        Raises:
            ValueError: If schema type cannot be detected
        """
        # Check for Pydantic models first
        if HAS_PYDANTIC:
            try:
                if isinstance(schema, type) and issubclass(schema, BaseModel):
                    return 'pydantic'
                elif isinstance(schema, BaseModel):
                    return 'pydantic'
            except TypeError:
                # Not a class that can be checked with issubclass
                pass

        # Check for marshmallow schemas
        if HAS_MARSHMALLOW:
            try:
                if isinstance(schema, Schema):
                    return 'marshmallow'
                elif isinstance(schema, type) and issubclass(schema, Schema):
                    return 'marshmallow'
                elif isinstance(schema, dict):  # Dict schemas are marshmallow
                    return 'marshmallow'
            except TypeError:
                # Not a class that can be checked with issubclass
                pass

        # Check for generic types like list[Model]
        if hasattr(schema, '__origin__') and hasattr(schema, '__args__'):
            # Handle generic types like list[Pet]
            if schema.__origin__ is list and len(schema.__args__) == 1:
                # Detect the inner type
                inner_type = schema.__args__[0]
                return self.detect_schema_type(inner_type)

        # Default to marshmallow for backwards compatibility
        if HAS_MARSHMALLOW:
            return 'marshmallow'

        raise ValueError(f'Cannot detect schema type for {type(schema)}')

    def create_adapter(
        self, schema: t.Any, schema_type: str | None = None, schema_name: str | None = None
    ) -> SchemaAdapter:
        """Create a schema adapter for the given schema.

        Arguments:
            schema: Schema object (can be a model or list[Model]/List[Model])
            schema_type: Optional schema type override
            schema_name: Optional schema name for dict schemas

        Returns:
            Schema adapter instance

        Raises:
            ValueError: If schema type is not supported
        """
        # Check if schema is a list type (list[Model] or List[Model])
        many = False
        inner_schema = schema

        if hasattr(schema, '__origin__') and hasattr(schema, '__args__'):
            if schema.__origin__ is list and len(schema.__args__) == 1:
                # Extract the inner model from list[Model] or List[Model]
                inner_schema = schema.__args__[0]
                many = True

        # Detect schema type from the inner schema
        if schema_type is None:
            schema_type = self.detect_schema_type(inner_schema)

        if schema_type not in self._adapters:
            available = ', '.join(self._adapters.keys())
            raise ValueError(
                f'Unsupported schema type: {schema_type}. ' f'Available types: {available}'
            )

        adapter_class = self._adapters[schema_type]
        return adapter_class(inner_schema, schema_name=schema_name, many=many)

    def get_available_types(self) -> list[str]:
        """Get list of available schema types.

        Returns:
            List of available schema type names
        """
        return list(self._adapters.keys())


# Global registry instance
registry = SchemaRegistry()



from __future__ import annotations

import typing as t

from .openapi import get_path_description
from .openapi import get_path_summary
from .types import ViewClassType
from .types import ViewFuncOrClassType
from .types import ViewFuncType
from .views import MethodView


def route_patch(cls):
    """A decorator to add a patched `add_url_rule` method for `APIFlask` and
    `APIBlueprint` objects.

    The patched `add_url_rule` method will create a view function if passed a
    view class, and then generate spec from it.

    *Version changed: 0.10.0*

    - Remove the `route` decorator, and move the logic into `add_url_rule`.

    *Version added: 0.5.0*
    """

    def record_spec_for_view_class(
        view_func: ViewFuncType, view_class: ViewClassType
    ) -> ViewFuncType:
        # when the user call add_url_rule multiple times for one view class,
        # we only need to extract info from view class once since it will
        # loop all the methods of the class.
        if hasattr(view_func, '_method_spec'):
            return view_func
        view_func._method_spec = {}
        if not hasattr(view_func, '_spec'):
            view_func._spec = {}
        if not view_class.methods:  # no methods defined
            return view_func
        for method_name in view_class.methods:  # type: ignore
            # method_name: ['GET', 'POST', ...]
            method = getattr(view_class, method_name.lower())
            # collect spec info from class attribute "decorators"
            if hasattr(view_func, '_spec') and view_func._spec != {}:
                if not hasattr(method, '_spec'):
                    method._spec = view_func._spec
                else:
                    for key, value in view_func._spec.items():
                        if value is not None and method._spec.get(key) is None:
                            method._spec[key] = value
            else:
                if not hasattr(method, '_spec'):
                    method._spec = {'no_spec': True}
            if not method._spec.get('summary'):
                method._spec['summary'] = get_path_summary(
                    method, f'{method_name.title()} {view_class.__name__}'
                )
                method._spec['generated_summary'] = True
            if not method._spec.get('description'):
                method._spec['description'] = get_path_description(method)
                method._spec['generated_description'] = True
            view_func._method_spec[method_name] = method._spec
        return view_func

    def add_url_rule(
        self,
        rule: str,
        endpoint: str | None = None,
        view_func: ViewFuncOrClassType | None = None,
        provide_automatic_options: bool | None = None,
        **options: t.Any,
    ):
        """Record the spec for view classes before calling the actual `add_url_rule` method.

        When calling this method directly, the `view_func` argument can be a view function or
        a view function created by `ViewClass.as_view()`. It only accepts a view class when
        using the route decorator on a view class.
        """
        if isinstance(view_func, type):
            # call as_view() for MethodView passed with @route
            if endpoint is None:
                endpoint = view_func.__name__
            view_func = view_func.as_view(endpoint)  # type: ignore

        if hasattr(view_func, 'view_class'):
            # view function created with MethodViewClass.as_view()
            view_class = view_func.view_class  # type: ignore
            if not issubclass(view_class, MethodView):
                # skip View-based class
                view_func._spec = {'hide': True}  # type: ignore
            else:
                # record spec for MethodView class
                if hasattr(self, 'enable_openapi') and self.enable_openapi:
                    view_func = record_spec_for_view_class(view_func, view_class)  # type: ignore

        super(cls, self).add_url_rule(
            rule,
            endpoint,
            view_func,
            provide_automatic_options=provide_automatic_options,
            **options,
        )

    cls.add_url_rule = add_url_rule
    return cls



import json

import openapi_spec_validator as osv
import pytest
from flask import request

from .schemas import Bar
from .schemas import Baz
from .schemas import Foo
from apiflask import APIBlueprint
from apiflask import Schema
from apiflask.commands import spec_command
from apiflask.fields import Integer


def test_spec(app):
    assert app.spec
    assert 'openapi' in app.spec


def test_spec_processor(app, client):
    @app.spec_processor
    def edit_spec(spec):
        assert spec['openapi'] == '3.0.3'
        spec['openapi'] = '3.0.2'
        assert app.title == 'APIFlask'
        assert spec['info']['title'] == 'APIFlask'
        spec['info']['title'] = 'Foo'
        return spec

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['openapi'] == '3.0.2'
    assert rv.json['info']['title'] == 'Foo'


def test_spec_processor_pass_object(app, client):
    app.config['SPEC_PROCESSOR_PASS_OBJECT'] = True

    class NotUsedSchema(Schema):
        id = Integer()

    @app.spec_processor
    def process_spec(spec):
        spec.title = 'Foo'
        spec.components.schema('NotUsed', schema=NotUsedSchema)
        return spec

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['info']['title'] == 'Foo'
    assert 'NotUsed' in rv.json['components']['schemas']
    assert 'id' in rv.json['components']['schemas']['NotUsed']['properties']


@pytest.mark.parametrize('spec_format', ['json', 'yaml', 'yml'])
def test_get_spec(app, spec_format):
    spec = app._get_spec(spec_format)

    if spec_format == 'json':
        assert isinstance(spec, dict)
    else:
        assert 'title: APIFlask' in spec


def test_get_spec_force_update(app):
    app._get_spec()

    @app.route('/foo')
    @app.output(Foo)
    def foo():
        pass

    spec = app._get_spec()
    assert '/foo' not in spec['paths']

    new_spec = app._get_spec(force_update=True)
    assert '/foo' in new_spec['paths']


def test_spec_bypass_endpoints(app):
    bp = APIBlueprint('foo', __name__, static_folder='static', url_prefix='/foo')
    app.register_blueprint(bp)

    spec = app._get_spec()
    assert '/static' not in spec['paths']
    assert '/foo/static' not in spec['paths']
    assert '/docs' not in spec['paths']
    assert '/openapi.json' not in spec['paths']
    assert '/docs/oauth2-redirect' not in spec['paths']


def test_spec_bypass_methods(app):
    class Foo:
        def bar(self):
            pass

    app.add_url_rule('/foo', 'foo', Foo().bar)

    spec = app._get_spec()
    assert '/foo' not in spec['paths']


def test_spec_attribute(app):
    spec = app._get_spec()

    @app.route('/foo')
    @app.output(Foo)
    def foo():
        pass

    assert '/foo' not in spec['paths']
    assert '/foo' in app.spec['paths']


def test_spec_schemas(app):
    @app.route('/foo')
    @app.output(Foo(partial=True))
    def foo():
        pass

    @app.route('/bar')
    @app.output(Bar(many=True))
    def bar():
        pass

    @app.route('/baz')
    @app.output(Baz)
    def baz():
        pass

    class Spam(Schema):
        id = Integer()

    @app.route('/spam')
    @app.output(Spam)
    def spam():
        pass

    class Ham(Schema):
        id = Integer()

    @app.route('/ham')
    @app.output(Ham)
    def ham():
        pass

    spec = app.spec
    assert len(spec['components']['schemas']) == 5
    assert 'FooUpdate' in spec['components']['schemas']
    assert 'Bar' in spec['components']['schemas']
    assert 'Baz' in spec['components']['schemas']
    assert 'Spam' in spec['components']['schemas']
    assert 'Ham' in spec['components']['schemas']


def test_servers_and_externaldocs(app):
    assert app.external_docs is None
    assert app.servers is None

    app.external_docs = {'description': 'Find more info here', 'url': 'https://docs.example.com/'}
    app.servers = [
        {'url': 'http://localhost:5000/', 'description': 'Development server'},
        {'url': 'https://api.example.com/', 'description': 'Production server'},
    ]

    rv = app.test_client().get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['externalDocs'] == {
        'description': 'Find more info here',
        'url': 'https://docs.example.com/',
    }
    assert rv.json['servers'] == [
        {'url': 'http://localhost:5000/', 'description': 'Development server'},
        {'url': 'https://api.example.com/', 'description': 'Production server'},
    ]


def test_default_servers(app):
    assert app.servers is None

    rv = app.test_client().get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    with app.test_request_context():
        assert rv.json['servers'] == [
            {
                'url': f'{request.url_root}',
            },
        ]


def test_default_servers_without_req_context(cli_runner):
    result = cli_runner.invoke(spec_command)
    assert 'openapi' in result.output
    assert 'servers' not in json.loads(result.output)


def test_auto_200_response(app, client):
    @app.get('/foo')
    def bare():
        pass

    @app.get('/bar')
    @app.input(Foo)
    def only_input():
        pass

    @app.get('/baz')
    @app.doc(summary='some summary')
    def only_doc():
        pass

    @app.get('/eggs')
    @app.output(Foo, status_code=204)
    def output_204():
        pass

    @app.get('/spam')
    @app.doc(responses={204: 'empty'})
    def doc_responses():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert '200' in rv.json['paths']['/foo']['get']['responses']
    assert '200' in rv.json['paths']['/bar']['get']['responses']
    assert '200' in rv.json['paths']['/baz']['get']['responses']
    assert '200' not in rv.json['paths']['/eggs']['get']['responses']
    assert '200' not in rv.json['paths']['/spam']['get']['responses']
    assert rv.json['paths']['/spam']['get']['responses']['204']['description'] == 'empty'


def test_sync_local_json_spec(app, client, tmp_path):
    app.config['AUTO_SERVERS'] = False

    local_spec_path = tmp_path / 'openapi.json'
    app.config['SYNC_LOCAL_SPEC'] = True
    app.config['LOCAL_SPEC_PATH'] = local_spec_path
    app.config['SPEC_FORMAT'] = 'json'

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)

    with open(local_spec_path) as f:
        spec_content = json.loads(f.read())
        assert spec_content == app.spec
        assert 'info' in spec_content
        assert 'paths' in spec_content


def test_sync_local_yaml_spec(app, client, tmp_path):
    app.config['AUTO_SERVERS'] = False

    local_spec_path = tmp_path / 'openapi.json'
    app.config['SYNC_LOCAL_SPEC'] = True
    app.config['LOCAL_SPEC_PATH'] = local_spec_path
    app.config['SPEC_FORMAT'] = 'yaml'

    rv = client.get('/openapi.json')
    assert rv.status_code == 200

    with open(local_spec_path) as f:
        spec_content = f.read()
        assert spec_content == str(app.spec)
        assert 'title: APIFlask' in spec_content


def test_sync_local_spec_no_path(app):
    app.config['SYNC_LOCAL_SPEC'] = True

    with pytest.raises(TypeError):
        app.spec



import typing as t

from apiflask import APIFlask, HTTPBasicAuth
from werkzeug.security import generate_password_hash, check_password_hash

app = APIFlask(__name__)
auth = HTTPBasicAuth()

users = {
    'userA': generate_password_hash('foo'),
    'userB': generate_password_hash('bar'),
}


@auth.verify_password
def verify_password(username: str, password: str) -> t.Union[str, None]:
    if username in users and check_password_hash(users[username], password):
        return username
    return None


@app.route('/')
@app.auth_required(auth)
def index():
    return f'Hello, {auth.current_user}'



from apiflask import APIFlask, APIBlueprint, Schema, abort
from apiflask.fields import Integer, String
from apiflask.validators import Length, OneOf

app = APIFlask(__name__)
hello_bp = APIBlueprint('hello', __name__)  # tag name will be "Hello"
pet_bp = APIBlueprint('pet', __name__)  # tag name will be "Pet"

# The default tag is the blueprint name in title form. If you need to set a custom tag name:
# hello_bp = APIBlueprint('hello', __name__, tag='NoHello')

# If you need to set the tag "description" or "externalDocs", just pass a dict:
# hello_bp = APIBlueprint('hello', __name__, tag={'name': 'Hello', 'description': '...'})
# pet_bp = APIBlueprint('pet', __name__, tag={'name': 'Pet', 'description': '...'})

pets = [
    {'id': 0, 'name': 'Kitty', 'category': 'cat'},
    {'id': 1, 'name': 'Coco', 'category': 'dog'},
    {'id': 2, 'name': 'Flash', 'category': 'cat'},
]


class PetIn(Schema):
    name = String(required=True, validate=Length(0, 10))
    category = String(required=True, validate=OneOf(['dog', 'cat']))


class PetOut(Schema):
    id = Integer()
    name = String()
    category = String()


@hello_bp.get('/')
def say_hello():
    return {'message': 'Hello!'}


@pet_bp.get('/pets/<int:pet_id>')
@pet_bp.output(PetOut)
def get_pet(pet_id):
    if pet_id > len(pets) - 1 or pets[pet_id].get('deleted'):
        abort(404)
    return pets[pet_id]


@pet_bp.get('/pets')
@pet_bp.output(PetOut(many=True))
def get_pets():
    return pets


@pet_bp.post('/pets')
@pet_bp.input(PetIn, location='json')
@pet_bp.output(PetOut, status_code=201)
def create_pet(json_data):
    pet_id = len(pets)
    json_data['id'] = pet_id
    pets.append(json_data)
    return pets[pet_id]


@pet_bp.patch('/pets/<int:pet_id>')
@pet_bp.input(PetIn(partial=True), location='json')
@pet_bp.output(PetOut)
def update_pet(pet_id, json_data):
    if pet_id > len(pets) - 1:
        abort(404)
    for attr, value in json_data.items():
        pets[pet_id][attr] = value
    return pets[pet_id]


@pet_bp.delete('/pets/<int:pet_id>')
@pet_bp.output({}, status_code=204)
def delete_pet(pet_id):
    if pet_id > len(pets) - 1:
        abort(404)
    pets[pet_id]['deleted'] = True
    pets[pet_id]['name'] = 'Ghost'
    return ''


app.register_blueprint(hello_bp)
app.register_blueprint(pet_bp)



from __future__ import annotations

import typing as t
from abc import abstractmethod
from typing import Protocol
from typing import runtime_checkable

if t.TYPE_CHECKING:  # pragma: no cover
    from flask.wrappers import Response  # noqa: F401
    from werkzeug.datastructures import Headers  # noqa: F401
    from _typeshed.wsgi import WSGIApplication  # noqa: F401
    from .fields import Field  # noqa: F401
    from .schemas import Schema  # noqa: F401
    from .security import HTTPBasicAuth  # noqa: F401
    from .security import HTTPTokenAuth  # noqa: F401
    from .security import APIKeyHeaderAuth  # noqa: F401
    from .security import APIKeyCookieAuth  # noqa: F401
    from .security import APIKeyQueryAuth  # noqa: F401
    from .exceptions import HTTPError  # noqa: F401
    from .views import View  # noqa: F401


DecoratedType = t.TypeVar('DecoratedType', bound=t.Callable[..., t.Any])
RequestType = t.TypeVar('RequestType')

ResponseBodyType = t.Union[
    str,
    bytes,
    t.List[t.Any],
    # Only dict is actually accepted, but Mapping allows for TypedDict.
    t.Mapping[str, t.Any],
    t.Iterator[str],
    t.Iterator[bytes],
    'Response',
]

ResponseStatusType = t.Union[str, int]
_HeaderName = str
_HeaderValue = t.Union[str, t.List[str], t.Tuple[str, ...]]
ResponseHeaderType = t.Union[
    t.Dict[_HeaderName, _HeaderValue],
    t.Mapping[_HeaderName, _HeaderValue],
    t.Sequence[t.Tuple[_HeaderName, _HeaderValue]],
    'Headers',
]
ResponseReturnValueType = t.Union[
    ResponseBodyType,
    t.Tuple[ResponseBodyType, ResponseHeaderType],
    t.Tuple[ResponseBodyType, ResponseStatusType],
    t.Tuple[ResponseBodyType, ResponseStatusType, ResponseHeaderType],
    'WSGIApplication',
]
SpecCallbackType = t.Callable[[t.Union[dict, str]], t.Union[dict, str]]
ErrorCallbackType = t.Callable[['HTTPError'], ResponseReturnValueType]

DictSchemaType = t.Dict[str, t.Union['Field', type]]
SchemaType = t.Union['Schema', t.Type['Schema'], DictSchemaType]
OpenAPISchemaType = t.Union['Schema', t.Type['Schema'], dict]
HTTPAuthType = t.Union[
    'HTTPBasicAuth', 'HTTPTokenAuth', 'APIKeyHeaderAuth', 'APIKeyCookieAuth', 'APIKeyQueryAuth'
]
TagsType = t.Union[t.List[str], t.List[t.Dict[str, t.Any]]]
ViewClassType = t.Type['View']
ViewFuncOrClassType = t.Union[t.Callable, ViewClassType]

ResponseObjectType = t.Dict[str, t.Union[str, t.Dict[str, t.Dict[str, t.Any]]]]
ResponsesObjectType = t.Dict[t.Union[int, str], ResponseObjectType]
ResponsesType = t.Union[t.List[int], t.Dict[int, str], ResponsesObjectType]

RouteCallableType = t.Union[
    t.Callable[..., ResponseReturnValueType],
    t.Callable[..., t.Awaitable[ResponseReturnValueType]],
]


class PaginationType(Protocol):
    page: int
    per_page: int
    pages: int
    total: int
    next_num: int
    has_next: bool
    prev_num: int
    has_prev: bool


class ViewFuncType(Protocol):
    _spec: t.Any
    _method_spec: t.Any


@runtime_checkable
class SecurityScheme(Protocol):
    name: str

    @abstractmethod
    def get_security_scheme(self) -> dict[str, t.Any]:
        """Get security scheme."""

        pass  # pragma: no cover
